@@ -8,7 +8,13 @@ import json
 import numbers
 
 from mathics.builtin.base import BoxConstructError
-from mathics.builtin.box.graphics import GraphicsBox, LineBox, PointBox, PolygonBox
+from mathics.builtin.box.graphics import (
+    GraphicsBox,
+    ArrowBox,
+    LineBox,
+    PointBox,
+    PolygonBox,
+)
 
 from mathics.builtin.colors.color_directives import _Color, RGBColor
 from mathics.builtin.drawing.graphics_internals import GLOBALS3D
@@ -694,12 +700,12 @@ currentlight=light(rgb(0.5,0.5,1), specular=red, (2,0,2), (2,2,2), (0,2,2));
         ]
 
 
-class Point3DBox(PointBox):
+class Arrow3DBox(ArrowBox):
     def init(self, *args, **kwargs):
-        super(Point3DBox, self).init(*args, **kwargs)
+        super(Arrow3DBox, self).init(*args, **kwargs)
 
     def process_option(self, name, value):
-        super(Point3DBox, self).process_option(name, value)
+        super(Arrow3DBox, self).process_option(name, value)
 
     def extent(self):
         result = []
@@ -715,6 +721,64 @@ class Point3DBox(PointBox):
                 coords.scale(boxscale)
 
 
+class Cylinder3DBox(_Graphics3DElement):
+    """
+    Internal Python class used when Boxing a 'Cylinder' object.
+    """
+
+    def init(self, graphics, style, item):
+        super(Cylinder3DBox, self).init(graphics, item, style)
+
+        self.edge_color, self.face_color = style.get_style(_Color, face_element=True)
+
+        if len(item.leaves) != 2:
+            raise BoxConstructError
+
+        points = item.leaves[0].to_python()
+        if not all(
+            len(point) == 3 and all(isinstance(p, numbers.Real) for p in point)
+            for point in points
+        ):
+            raise BoxConstructError
+
+        self.points = [Coords3D(graphics, pos=point) for point in points]
+        self.radius = item.leaves[1].to_python()
+
+    def to_asy(self):
+        if self.face_color is None:
+            face_color = (1, 1, 1)
+        else:
+            face_color = self.face_color.to_js()
+
+        rgb = f"rgb({face_color[0]}, {face_color[1]}, {face_color[2]})"
+        return "".join(
+            f"draw(surface(cylinder({tuple(coord.pos()[0])}, {self.radius}, {self.height})), {rgb});"
+            for coord in self.points
+        )
+
+    def extent(self):
+        result = []
+        # FIXME: instead of `coords.add(±self.radius, ±self.radius, ±self.radius)` we should do:
+        # coords.add(transformation_vector.x * ±self.radius, transformation_vector.y * ±self.radius, transformation_vector.z * ±self.radius)
+        result.extend(
+            [
+                coords.add(self.radius, self.radius, self.radius).pos()[0]
+                for coords in self.points
+            ]
+        )
+        result.extend(
+            [
+                coords.add(-self.radius, -self.radius, -self.radius).pos()[0]
+                for coords in self.points
+            ]
+        )
+        return result
+
+    def _apply_boxscaling(self, boxscale):
+        # TODO
+        pass
+
+
 class Line3DBox(LineBox):
     def init(self, *args, **kwargs):
         super(Line3DBox, self).init(*args, **kwargs)
@@ -736,6 +800,27 @@ class Line3DBox(LineBox):
                 coords.scale(boxscale)
 
 
+class Point3DBox(PointBox):
+    def init(self, *args, **kwargs):
+        super(Point3DBox, self).init(*args, **kwargs)
+
+    def process_option(self, name, value):
+        super(Point3DBox, self).process_option(name, value)
+
+    def extent(self):
+        result = []
+        for line in self.lines:
+            for c in line:
+                p, d = c.pos()
+                result.append(p)
+        return result
+
+    def _apply_boxscaling(self, boxscale):
+        for line in self.lines:
+            for coords in line:
+                coords.scale(boxscale)
+
+
 class Polygon3DBox(PolygonBox):
     def init(self, *args, **kwargs):
         self.vertex_normals = None
@@ -805,9 +890,11 @@ class Sphere3DBox(_Graphics3DElement):
 # FIXME: GLOBALS3D is a horrible name.
 GLOBALS3D.update(
     {
-        "System`Polygon3DBox": Polygon3DBox,
+        "System`Arrow3DBox": Arrow3DBox,
+        "System`Cylinder3DBox": Cylinder3DBox,
         "System`Line3DBox": Line3DBox,
         "System`Point3DBox": Point3DBox,
+        "System`Polygon3DBox": Polygon3DBox,
         "System`Sphere3DBox": Sphere3DBox,
     }
 )
